#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Written by KP_Hong <kphong16@daum.net>
2021, KP_Hong
"""

import pandas as pd
import numpy as np
from pandas import Series, DataFrame

from datetime import datetime
from datetime import date
from functools import wraps

from .genfunc import (
    PY, 
    limited,
    is_scalar,
    is_iterable,
    )
from .index import(
    RangeIndex,
    DateIndex,
    date_range,
    str_to_date,
    )
from .account import (
    Account,
    Merge
    )

__all__ = ['Loan', 'Merge_loans']

class Loan(object):
    """
    Parameters
    ----------
    index       : Index, basic index class
    idxfn       : Index, financial index class
    mtrt        : int, maturity of loan
    amt_ntnl    : int or float, notional amount
    rate_IR     : float, interest rate
    rate_fee    : float, initial fee rate
    rate_fob    : float, fee on balance which is not withdrawed
    rate_arng   : float, arangement fee rate
    cycle_IR    : int, interest payment cycle(months), 1 is default
    title       : str, name of loan
    byname      : str, byname of loan
    rnk         : int, rank of loan
    **kwargs    : others
    
    Class Attributes
    ----------------
    dct         : dictionary of loans
    index       : basic index class
    idxfn       : financial index class
    mtrt        : maturity of loan
    rate_arng   : arangement fee rate
    ttl_ntnl    : total amount of notional amounts
    ttl_intl    : total amount of initial amounts
    amt_arng    : amount of arangement fee
    cycle_IR    : interest payment cycle(months), 1 is default
    
    Class Methods
    -------------
    rnk(reverse=False)  : return sorted list of rank
    is_repaid_all()     : return boolean whether loans are repaid or not
    mrgloans()          : return merge of loans
    by_rnk(rnk)         : find loan by rnk
    allin_ttl()         : total all-in rate
    
    Instance Attributes
    -------------------
    dct_item    : dictionary of loan items(ntnl, IR, etc.)
    mrgitems    : merge of item dictionary
    _df         : dataframe of mrgitems
    df          : dataframe of mrgitems
    index       : basic index class
    idxfn       : financial index class
    mtrt        : maturity of loan
    
    amt_ntnl    : notional amount
    amt_intl    : initial amount
    rate_IR     : interest rate
    rate_fee    : initial fee rate
    rate_fob    : fee on balance which is not withdrawed
    
    title       : name of loan
    byname      : byname of loan
    rnk         : rank of loan
    ntnl        : account of notional amount
    
    IR          : account of interest rate
    IR.rate     : interest rate
    IR.cycle    : interest payment cycle(months), 1 is default
    IR.rate_cycle : interest rate on a cycle
    IR.amt_year : interest rate amount on a year
    IR.amt_cycle  : interest rate amount on a cycle
    
    fee         : account of fee
    fee.rate    : initial fee rate
    fee.amt     : fee amount
    fob         : account of fee on balance
    fob.rate    : fob rate
    fob.cycle   : cycle of fob
    fob.rate_cycle : rate of fob on cycle
    is_wtdrbl   : whether it is withdrawable
    is_repaid   : whether it is repaid
    
    Instance Methods
    ----------------
    set_wtdrbl_intldate(date, basedate=None)
    setback_wtdrbl_mtrt(date)
    set_wtdrbl_false()
    set_repaid(date)
    IRamt_topay(idxno)      : calculate IR amount to pay
    fobamt_topay(idxno)     : calculate fob amount to pay
    wtdrw(idxno, amt, acc)  : withdraw amount from the notional and send to the account
    ntnl_bal_end(idxno)     : notional balance that is not repayed on the index date
    amt_rpy_exptd(idxno)    : notional amount which repayment date has arrived
    ntnl_out_rsdl(idxno)    : residual loan notional amount that is withdrawble
    amt_repay(idxno, amt)   : input the amount and return the amount within notional balance end
    """
    
    def __new__(cls, *args,**kwargs):
        cls.key_all = [key for key in kwargs.keys()]
        cls.key_essential = ['index', 'idxfn', 'mtrt', 'amt_ntnl', 'rate_IR']
        cls.key_optional = []
        cls.dct = {}
        cls._rnk = []
        cls.cycle_IR = 1
        cls.no_loan_dev = 0
        
        # Triple new
        if len(args) == 3:
            if isinstance(args[0], DateIndex):
                cls.index = args[0]
                cls.idxfn = cls.index
                cls.mtrt = len(args[0])
            elif isinstance(args[0], int):
                cls.mtrt = args[0]
                cls.index = RangeIndex(args[0])
                cls.idxfn = cls.index
                
            if isinstance(args[1], (int, float)):
                amt_ntnl = args[1]
            else:
                raise ValueError("amt_ntnl should be int or float.")
            
            if isinstance(args[2], float):
                rate_IR     = args[2]
            else:
                raise ValueError("rate_IR should be float")

            kwargs = {key: item for key, item in kwargs.items() if key not in cls.key_essential}
            
            return cls._triple_new(amt_ntnl, rate_IR, **kwargs)
            
        # Keyward arguments
        if len(args) == 0:
            for key, item in kwargs.items():
                setattr(cls, key, item)
            return cls._kwargs_new(**kwargs)
            
            
    @classmethod
    def _triple_new(cls, amt_ntnl, rate_IR, **kwargs):
        
        data = {
            'index'     : cls.index,
            'idxfn'     : cls.idxfn,
            'mtrt'      : cls.mtrt,
            'amt_ntnl'  : amt_ntnl,
            'rate_IR'   : rate_IR,
        }
        data.update(kwargs)
        result_one_val = cls._onevalue_kwargs_new(**data)
        cls._after_calculation()
        cls.this = result_one_val
        return cls

    
    @classmethod
    def _kwargs_new(cls, **kwargs):        
        if 'index' in kwargs:
            cls.index = kwargs.pop('index')
            if 'idxfn' in kwargs:
                cls.idxfn = kwargs.pop('idxfn')
            else:
                cls.idxfn = cls.index
            if 'mtrt' in kwargs:
                cls.mtrt = kwargs.pop('mtrt')
            else:
                cls.mtrt = len(cls.idxfn)
        if 'mtrt' in kwargs:
            cls.mtrt = kwargs.pop('mtrt')
            cls.index = RangeIndex(cls.mtrt)
            cls.idxfn = cls.index
        if 'rate_arng' in kwargs:
            cls.rate_arng = kwargs.pop('rate_arng')
        
        try:
            amt_ntnl = kwargs['amt_ntnl']
        except:
            ValueError("'amt_ntnl' is essential")
        
        if is_iterable(amt_ntnl):
            cls.number_of_loans = len(amt_ntnl)
            
            for no, _amt_ntnl in enumerate(amt_ntnl):
                data = {key: item[no] for key, item in kwargs.items()}
                cls._onevalue_kwargs_new(**data)
                
            cls._after_calculation()
            return cls
            
        elif not is_iterable(amt_ntnl):
            cls.number_of_loans = 1
            
            data = kwargs.copy()
            result_one_val = cls._onevalue_kwargs_new(**data)
            cls._after_calculation()
            cls.this = result_one_val
            return cls
    
    @classmethod
    def _after_calculation(cls):
        cls.ttl_ntnl = sum([item.amt_ntnl for item in cls.dct.values()])
        if 'amt_intl' in cls.key_all:
            cls.ttl_intl = sum([item.amt_intl for item in cls.dct.values()])
        if 'rate_arng' in cls.key_all:
            cls.amt_arng = cls.rate_arng * cls.ttl_ntnl
    
    @classmethod
    def _onevalue_kwargs_new(cls, **data):
        result = object.__new__(cls)
        result.dct_item = {}
        
        result.index = cls.index
        result.idxfn = cls.idxfn
        result.mtrt = cls.mtrt
        
        result._is_wtdrbl = False
        result._is_repaid = False
        
        # Rank
        if 'rnk' in data:
            result.rnk = data.pop('rnk')
        else:
            result.rnk = 0
        cls._rnk.append(result.rnk)
        
        # Notional amount
        try:
            amt_ntnl = data.pop('amt_ntnl')
        except:
            ValueError("'amt_ntnl' is essential")
        
        if 'amt_intl' in data:
            amt_intl = data['amt_intl']
        else:
            amt_intl = amt_ntnl
            
        _ntnl = Account(cls.index)
        #_ntnl.subscd(cls.idxfn[0], amt_intl)
        #_ntnl.addscd(cls.idxfn[-1], amt_ntnl)
        result.amt_ntnl = amt_ntnl
        result.amt_intl = amt_intl
        result.ntnl = _ntnl
        result.dct_item['ntnl'] = _ntnl
        
        # Interest rate
        if 'rate_IR' in data:
            rate_IR = data.pop('rate_IR')
            if 'cycle_IR' in data:
                cls.cycle_IR = data.pop('cycle_IR')

            _IR = Account(cls.index)
            _IR.rate = rate_IR
            _IR.cycle = cls.cycle_IR
            _IR.rate_cycle = rate_IR * cls.cycle_IR / 12
            _IR.amt_year = rate_IR * amt_ntnl
            _IR.amt_cycle = _IR.rate_cycle * amt_ntnl
            result.rate_IR = rate_IR
            result.IR = _IR
            result.dct_item['IR'] = _IR
        
        # Fee
        if 'rate_fee' in data:
            rate_fee = data.pop('rate_fee')
            
            _fee = Account(cls.index)
            _fee.rate = rate_fee
            _fee.amt = rate_fee * amt_ntnl
            result.rate_fee = rate_fee
            result.fee = _fee
            result.dct_item['fee'] = _fee
            
        # Fob(Fee on balance)
        if 'rate_fob' in data:
            rate_fob = data.pop('rate_fob')
            
            _fob = Account(cls.index)
            _fob.rate = rate_fob
            _fob.cycle = cls.cycle_IR
            _fob.rate_cycle = rate_fob * cls.cycle_IR / 12
            result.rate_fob = rate_fob
            result.fob = _fob
            result.dct_item['fob'] = _fob
            
        # Allin
        cst_year = 0
        if 'rate_IR' in cls.key_all:
            cst_year += amt_ntnl * rate_IR
        if 'rate_fee' in cls.key_all:
            cst_year += amt_ntnl * rate_fee / cls.mtrt * 12
        if amt_ntnl == 0 or amt_ntnl is None:
            result.allin = 0
        else:
            result.allin = cst_year / amt_ntnl
        
        # Others
        if 'title' in data:
            title = data.pop('title')
            result.title = title
        else:
            title = f"tr{cls.no_loan_dev}"
            cls.no_loan_dev += 1
            result.title = title
        
        if 'byname' in data:
            byname = data.pop('byname')
            result.byname = byname
            
        for key, item in data.items():
            setattr(result, key, item)
            
        cls.dct[title] = result
        setattr(cls, title, result)
        return result
    

    @classmethod
    def rnk(cls, reverse=False):
        if reverse == False:
            return sorted(cls._rnk, reverse=reverse)
        elif reverse == True:
            return sorted(cls._rnk, reverse=reverse)
            
    @classmethod
    def is_repaid_all(cls):
        for key, item in cls.dct.items():
            if item.is_repaid == False:
                return False
        return True
        
    @classmethod
    def mrgloans(cls):
        return Merge_loans(cls.dct)
        
    @classmethod
    def by_rnk(cls, rnk):
        for key, item in cls.dct.items():
            if item.rnk == rnk:
                return item
                
    @classmethod
    def allin_ttl(cls):
        tmpamt = 0
        for key, item in cls.dct.items():
            tmpamt += item.allin * item.amt_ntnl
        return tmpamt / cls.ttl_ntnl
        
    @property   
    def mrgitems(self):
        return Merge(self.dct_item)
            
    @property
    def _df(self):
        return self.mrgitems._df
        
    @property
    def df(self):
        return self.mrgitems.df
    
    # Set loan withdrawable
    @property
    def is_wtdrbl(self):
        return self._is_wtdrbl
    @is_wtdrbl.setter
    def is_wtdrbl(self, value):
        self._is_wtdrbl = value
        
    @property
    def keys(self):
        return list(self.__dict__.keys())
        
    def set_wtdrbl_intldate(self, date, basedate=None):
        """If the date is the initial date, then set is_wtdrbl True."""
        if basedate is None:
            basedate = self.idxfn[0]
        if date == basedate:
            self.is_wtdrbl = True
            
    def setback_wtdrbl_mtrt(self, date):
        """If the date is a maturity date, then set back is_wtdrbl False."""
        if date == self.idxfn[-1]:
            self.set_wtdrbl_false()
            
    def set_wtdrbl_false(self):
        """Set is_wtdrbl False"""
        self.is_wtdrbl = False
        
    # Set loan repaid all
    @property
    def is_repaid(self):
        return self._is_repaid
    @is_repaid.setter
    def is_repaid(self, value):
        self._is_repaid = value
    
    def set_repaid(self, date):
        if self.is_wtdrbl == False:
            return
        if -self.ntnl.bal_end[date] > 0:
            return
        if (self.ntnl.rsdl_out_cum[date] - max(self.ntnl.rsdl_in_cum[date], 0)) > 0:
            return
        self.is_repaid = True
        
    # Calculate IR amount to pay
    def IRamt_topay(self, idxno):
        IRamt = -self.ntnl.bal_strt[idxno] * self.IR.rate_cycle
        return IRamt
        
    def fobamt_topay(self, idxno):
        if 'rate_fob' in self.key_all:
            fobamt = self.ntnl_out_rsdl(idxno) * self.fob.rate_cycle
            return fobamt
        raise ValueError('fob is not created')

    # Reset IR rate
    def reset_IR(self, new_IR):
        self.IR.rate = new_IR
        self.IR.rate_cycle = self.IR.rate * self.IR.cycle / 12
    
    # Withdraw loan
    def wtdrw(self, idxno, amt, acc):
        """Withdraw the amount to the account within the withdrawable balance limit."""
        if not self.is_wtdrbl:
            return 0
        if self.is_repaid:
            return 0
        
        amt_wtdrw = limited(amt,
                            upper = [self.ntnl.rsdl_out_cum[idxno]],
                            lower = [0])
        if amt_wtdrw > 0:
            self.ntnl.send(idxno, amt_wtdrw, acc)
        return amt_wtdrw
    
    # Repayment
    def ntnl_bal_end(self, idxno):
        """Notional balance that is not repayed."""
        return -self.ntnl.bal_end[idxno]
    
    def amt_rpy_exptd(self, idxno):
        """Notional amount which repayment date has arrived."""
        amt_rpy = limited(self.ntnl.rsdl_in_cum[idxno],
                          upper = [self.ntnl_bal_end(idxno)],
                          lower = [0])
        return amt_rpy
        
    def ntnl_out_rsdl(self, idxno):
        """Residual loan notional amount that is withdrawble."""
        amt_rsdl = limited(self.ntnl.rsdl_out_cum[idxno],
                           lower=[0])
        return amt_rsdl
    
    def amt_repay(self, idxno, amt):
        amtrpy = limited(amt,
                         upper = [self.ntnl_bal_end(idxno)],
                         lower = [0])
        return amtrpy
        
    def __repr__(self):
        """
        Return a string representation for this object.
        """
        repr_smry =(f"{'Title':<10}: {self.title}\n" + 
                    f"{'Notional':<10}: {self.amt_ntnl:,.0f}\n"
                    )
        if 'IR' in self.__dict__:
            repr_tmp = f"{'IR':<10}: {self.rate_IR*100:.1f}%\n"
            repr_smry += repr_tmp
        
        if 'fee' in self.__dict__:
            repr_tmp = f"{'Fee':<10}: {self.rate_fee:.1f}%\n"
            repr_smry += repr_tmp
        
        if 'fob' in self.__dict__:
            repr_tmp = f"{'Fob':<10}: {self.rate_fob:.1f}%\n"
            repr_smry += repr_tmp
        
        repr_smry += str(self.__class__)
        
        return repr_smry
        
        
class Merge_loans(Merge):
    @property
    def ntnl(self):
        tmp_dct = {key:val.ntnl for key, val in self.dct.items()}
        rslt_acc = Merge(tmp_dct)
        return rslt_acc
        
    @property
    def IR(self):
        tmp_dct = {key:val.IR for key, val in self.dct.items()}
        rslt_acc = Merge(tmp_dct)
        return rslt_acc
        
    @property
    def fee(self):
        tmp_dct = {key:val.fee for key, val in self.dct.items() if 'fee' in val.__dict__.keys()}
        rslt_acc = Merge(tmp_dct)
        return rslt_acc
        
    @property
    def fob(self):
        tmp_dct = {key:val.fob for key, val in self.dct.items() if 'fob' in val.__dict__.keys()}
        rslt_acc = Merge(tmp_dct)
        return rslt_acc    
    
    
    def __repr__(self):
        """Return a string representation for this object."""
        repr_key = []
        for key in self.dct.keys():
            repr_key.append(key)
            
        return f"Merge_of_Loans: {repr_key}"
                
        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Written by KP_Hong <kphong16@daum.net>
2021, KP_Hong

CAFLE
-----
"cafle" is a package that can be used to estimate the cash flow of financial investments. "Account" and "Loan" modules on "cafle" make it easier to estimate the cash flow of the financial investments.

Financial planners and investors in charge of investment projects often estimate the cash flow of investment projects to review the adequacy of investment. This package can be used where various financial models are needed, from large projects such as some real estate development projects, and infrastructure development projects to individual financial plans.

In many cases, it takes more than a few weeks to create a financial model with Excel, but it also takes a considerable amount of time to modify or create a new model according to investment requirements. If the size of the model increases above a certain level, it is not easy to interpret or modify it.

"cafle" will greatly reduce the time spent using Excel.
"""

from .genfunc import *
from .index import *
from .account import *
from .loan import *
from .write import *
from .setaccount import *
from .standarddata import *




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Written by KP_Hong <kphong16@daum.net>
2021, KP_Hong

ACCOUNT

Modules
-------


Attributes
----------


Methods
-------


"""

import pandas as pd
import numpy as np
from pandas import Series, DataFrame

from datetime import datetime
from datetime import date
from functools import wraps

from .genfunc import *
from .genfunc import listwrapper
from .index import (
    RangeIndex,
    DateIndex,
    date_range,
    str_to_date,
    )

__all__ = ['Account', 'Merge',
           'set_acc', 'set_once', 'set_scd', 'set_rate']

DFCOL      = ['scd_in', 'scd_in_cum', 'scd_out', 'scd_out_cum', 
                           'bal_strt', 'amt_in', 'amt_in_cum', 
                           'amt_out', 'amt_out_cum', 'bal_end',
                           'rsdl_in_cum', 'rsdl_out_cum']
DFCOL_smry = ['bal_strt', 'amt_in', 'amt_out', 'bal_end']
DFCOL_inpt = ['scd_in', 'scd_out', 'amt_in', 'amt_out']
JNLCOL     = ['amt_in', 'amt_out', 'rcvfrm', 'payto', 'note']

class Account(object):
    """
    Parameters
    ----------
    index : index class
    title : str, name of account
    balstrt : int, float, default 0, start balance of account
    
    Attributes
    ----------
    index : index class
    title : str, name of account
    balstrt : start balance of account
    DFCOL : columns of the dataframe
    DFCOL_smry : summary columns of the dataframe
    JNLCOL : columns of the journal
        
    Methods
    -------
    <input data>
    addscd(index, amt) : add the amt on the 'scd_in' column of the dataframe.
    subscd(index, amt) : add the amt on the 'scd_out' column of the dataframe.
    addamt(index, amt, rcvfrm=None, note="add_amt")
        add the amt on the 'amt_in' column of the dataframe.
    subamt(index, amt, payto=None, note="sub_amt")
        add the amt on the 'amt_out' column of the dataframe.
    iptamt(index, amt, rcvfrm=None, payto=None, note=None):
        if amt is positive, apply addamt, else apply subamt.
    iptjnl(index, amt_in, amt_out, rcvfrm=None, payto=None, note=None)
        add the amt on the journal.
    
    <output data>
    df : return the summarised dataframe
    _df : return the total dataframe
    jnl : return the journal dataframe
    scd_in[slice] : return the value on the scd_in column.
    scd_in_cum[slice] : return the value on the scd_in_cum column.
    scd_out[slice] : return the value on the scd_out column.
    scd_out_cum[slice] : return the value on the scd_out_cum column.
    bal_strt[slice] : return the value on the bal_strt column.
    amt_in[slice] : return the value on the amt_in column.
    amt_in_cum[slice] : return the value on the amt_in_cum column.
    amt_out[slice] : return the value on the amt_out column.
    amt_out_cum[slice] : return the value on the amt_out_cum column.
    bal_end[slice] : return the value on the bal_end column.
    rsdl_in_cum[slice] : return the value on the rsdl_in_cum column.
    rsdl_out_cum[slice] : return the value on the rsdl_out_cum column.
    
    <etc.>
    amt_rqrd_excs(index, rqrdamt, minunit=100)
        return an additional amount compared to the account balance.
    send(index, amt, account, note=None)
        transfer the amount from this account to the opponent account.
    """
    
    _index = None
    
    def __init__(self,
                 data=None,
                 index=None,
                 **kwargs
                 ):
        """             
        DFCOL      = ['scd_in', 'scd_in_cum', 'scd_out', 'scd_out_cum', 
                           'bal_strt', 'amt_in', 'amt_in_cum', 
                           'amt_out', 'amt_out_cum', 'bal_end',
                           'rsdl_in_cum', 'rsdl_out_cum']
        DFCOL_smry = ['bal_strt', 'amt_in', 'amt_out', 'bal_end']
        DFCOL_inpt = ['scd_in', 'scd_out', 'amt_in', 'amt_out']
        JNLCOL     = ['amt_in', 'amt_out', 'rcvfrm', 'payto', 'note']
        """             
        
        if isinstance(data, (DateIndex, RangeIndex)) and index is None:
            index = data
            data = None
        
        if index is None:
            if isinstance(data, DataFrame):
                _index = DataFrame.index
            else:
                _index = self.__class__._index
        elif isinstance(index, RangeIndex):
            _index = index
        elif isinstance(index, DateIndex):
            _index = index
        #elif isinstance(index, list):
        #    _index = list(index)
        #elif is_iterable(index):
        #    _index = list(index)
        
        if _index is None:
            _df = DataFrame(columns=DFCOL)
        else:
            _df = DataFrame(np.zeros([len(_index), len(DFCOL)]),
                            columns = DFCOL,
                            index = _index)
            
        if isinstance(data, DataFrame):
            for key in data.columns:
                if key in DFCOL_inpt:
                    _df.loc[data.index, key] = data[key]
                else:
                    raise ValueError
        
        _jnl = DataFrame(columns = JNLCOL)
        
        self.index = _index
        self._df = _df
        self._jnl = _jnl
        
        for key, item in kwargs.items():
            kwargsitem = ['title', 'byname']
            if key in kwargsitem:
                setattr(self, key, item)
            elif key not in kwargsitem:
                setattr(self, key, item)
        
        self._intlz()
        self._cal_bal()
        
    def _intlz(self):
        # Initial Setting Function
        self._set_outputfunc()
        
        
    # Calculate Data Balance
    def _cal_bal(self):
        if self.index is None:
            return None
        
        self._df.scd_in_cum = self._df.scd_in.cumsum()
        self._df.scd_out_cum = self._df.scd_out.cumsum()
        self._df.amt_in_cum = self._df.amt_in.cumsum()
        self._df.amt_out_cum = self._df.amt_out.cumsum()
        
        # Calculate account balance
        for i, idx in enumerate(self.index):
            if i > 0:
                self._df.loc[idx, 'bal_strt'] = self._df.loc[idxpst, 'bal_end']
            self._df.loc[idx, 'bal_end'] = self._df.loc[idx, 'bal_strt'] \
                                         + self._df.loc[idx, 'amt_in'] \
                                         - self._df.loc[idx, 'amt_out']
            idxpst = idx
            
        # Calculate the residual of scheduled amounts
        self._df.loc[:, 'rsdl_in_cum'] = self._df.loc[:, 'scd_in_cum'] \
                                       - self._df.loc[:, 'amt_in_cum']
        self._df.loc[:, 'rsdl_out_cum'] = self._df.loc[:, 'scd_out_cum'] \
                                        - self._df.loc[:, 'amt_out_cum']
                                        
    # Input Data
    @listwrapper
    def addscd(self, idxval, amt):
        """
        Add the amount on the 'scd_in' column of dataframe.
        
        Parameters
        ----------
        idxval : index, list
        amt : int, float, list
        
        Returns
        -------
        None
        """
        if isinstance(idxval, int): 
            idxval = self.index[idxval]
        elif isinstance(idxval, str):
            idxval = str_to_date(idxval)
        
        self._df.loc[idxval, 'scd_in'] += amt
        self._cal_bal()
    
    @listwrapper
    def subscd(self, idxval, amt):
        """
        Add the amount on the 'scd_out' column of dataframe.
        
        Parameters
        ----------
        idxval : index, list
        amt : int, float, list
        
        Returns
        -------
        None
        """
        if isinstance(idxval, int): 
            idxval = self.index[idxval]
        elif isinstance(idxval, str):
            idxval = str_to_date(idxval)
            
        self._df.loc[idxval, 'scd_out'] += amt
        self._cal_bal()
        
    @listwrapper
    def addamt(self, idxval, amt, rcvfrm=None, note="add_amt"):
        """
        Add the amount on the 'amt_in' column of dataframe.
        
        Parameters
        ----------
        idxval : index, list
        amt : int, float, list
        rcvfrm : str, default None
        note : str, default "add_amt"
        
        Returns
        -------
        None
        """
        if isinstance(idxval, int): 
            idxval = self.index[idxval]
        elif isinstance(idxval, str):
            idxval = str_to_date(idxval)
            
        if amt == 0:
            return    
        self.iptjnl(idxval, amt, 0, rcvfrm, None, note)
        self._df.loc[idxval, 'amt_in'] += amt
        self._cal_bal()
        
    @listwrapper
    def subamt(self, idxval, amt, payto=None, note="sub_amt"):
        """
        Add the amount on the 'amt_out' column of dataframe.
        
        Parameters
        ----------
        idxval : index, list
        amt : int, float, list
        payto : str, default None
        note : str, default "sub_amt"
        
        Returns
        -------
        None
        """
        if isinstance(idxval, int): 
            idxval = self.index[idxval]
        elif isinstance(idxval, str):
            idxval = str_to_date(idxval)
            
        if amt == 0:
            return
        self.iptjnl(idxval, 0, amt, None, payto, note)
        self._df.loc[idxval, 'amt_out'] += amt
        self._cal_bal()
        
    @listwrapper
    def iptamt(self, idxval, amt, rcvfrm=None, payto=None, note=None):
        """
        If the amount is positive, apply the addamt, else apply the subamt
        
        Parameters
        ----------
        idxval : index, list
        amt : int, float, list
        rcvfrm : str, default None
        payto : str, default None
        note : str, default "sub_amt"
        
        Returns
        -------
        None
        """
        if isinstance(idxval, int): 
            idxval = self.index[idxval]
        elif isinstance(idxval, str):
            idxval = str_to_date(idxval)
            
        if amt == 0:
            return
        if amt > 0:
            if rcvfrm is None:
                rcvfrm = "add_amt"
            self.addamt(idxval, amt, rcvfrm, note)
        else:
            if payto is None:
                payto = "sub_amt"
            self.subamt(idxval, -amt, payto, note)
        
    def iptjnl(self, idxval, amt_in, amt_out, rcvfrm=None, payto=None, note=None):
        """
        Add the amount on the journal.
        
        Parameters
        ----------
        idxval : index
        amt_in : int, float
        amt_out : int, float
        rcvfrm : str, default None
        payto : str, default None
        note : str, default None
        """
        if isinstance(idxval, int): 
            idxval = self.index[idxval]
        elif isinstance(idxval, str):
            idxval = str_to_date(idxval)
            
        tmpjnl = DataFrame([[amt_in, amt_out, rcvfrm, payto, note]], 
                           columns=JNLCOL, index=[idxval])
        self._jnl = pd.concat([self._jnl, tmpjnl])
    
    def __getattr__(self, attr):
        return self.__dict__[attr]
        
        
    # Output Data
    @property
    def df(self):
        """
        Return the summarised dataframe
        Return columns : DFCOL_smry
            ['bal_strt', 'amt_in', 'amt_out', 'bal_end']
        """
        return self._df.loc[:, DFCOL_smry]
    
    @property
    def jnl(self):
        """
        Return the journal dataframe
        Return columns : JNLCOL
            ['amt_in', 'amt_out', 'rcvfrm', 'payto', 'note']
        """
        return self._jnl
    
    # Decorator
    class getattr_dfcol:
        """
        Decorator
        Get a class name and use the class name as the column of dataframe.
        """
        def __call__(self, cls):
            def init(self, spristnc):
                self.spristnc = spristnc
                self.colname = cls.__name__
            cls.__init__ = init
            
            def getitem(self, val):
                """
                If val is an integer, return the data which is in index[val].
                If val is a date, return the data which is on the date.
                If val is a string, get the date data and return the data 
                    which is on the date.
                
                Parameters
                ----------
                val: int, slice, date, date like string, 
                    ex) 0, 1:3, datetime.date(2021, 4, 30), "2021-04", "2021-04-30"
                
                Return
                ------
                data from dataframe
                Array of data from dataframe
                
                Examples
                --------
                >>> idx = DateIndex("2021.01", "2021.12")
                >>> acc = Account(idx, "loan")
                >>> acc.addscd(idx[0], 1000)
                >>> acc.subscd(idx[5], 800)
                >>> acc.addamt(idx[1], 500, "acc_oprtg", "amount acc oprtg")
                >>> acc.scd_in[idx[3]]
                    0.0
                >>> acc.scd_in[3]
                    0.0
                >>> acc.scd_in[0:3]
                    2021-01-31    1000.0
                    2021-02-28       0.0
                    2021-03-31       0.0
                    Name: scd_in, dtype: float64
                >>> acc.scd_in[datetime.date(2021, 3, 31)]
                    0.0
                >>> acc.scd_in["2021.04"]
                    2021-04-30    0.0
                    Name: scd_in, dtype: float64
                >>> acc.scd_in["2021"]
                    2021-01-31    1000.0
                    2021-02-28       0.0
                    2021-03-31       0.0
                    2021-04-30       0.0
                    2021-05-31       0.0
                    2021-06-30       0.0
                    2021-07-31       0.0
                    2021-08-31       0.0
                    2021-09-30       0.0
                    2021-10-31       0.0
                    2021-11-30       0.0
                    Name: scd_in, dtype: float64
                """
                if isinstance(val, date):
                    return self.spristnc._df.loc[val, self.colname]
                elif isinstance(val, int):
                    val = self.spristnc.index[val]
                    return self.spristnc._df.loc[val, self.colname]
                elif isinstance(val, str):
                    val = str_to_date(val)
                    return self.spristnc._df.loc[val, self.colname]
                elif isinstance(val, slice):
                    if isinstance(val.start, str):
                        new_start = str_to_date(val.start)
                    elif isinstance(val.start, int):
                        new_start = self.spristnc.index[val.start]
                    elif isinstance(val.start, date):
                        new_start = val.start
                        
                    if isinstance(val.stop, str):
                        new_stop = str_to_date(val.stop)
                    elif isinstance(val.stop, int):
                        new_stop = self.spristnc.index[val.stop]
                    elif isinstance(val.stop, date):
                        new_stop = val.stop
                        
                    new_slice = slice(new_start, new_stop)
                    return self.spristnc._df.loc[new_slice, self.colname]
            
            cls.__getitem__ = getitem
            
            return cls
        
    @getattr_dfcol()
    class scd_in:
        pass
    @getattr_dfcol()
    class scd_in_cum:
        pass
    @getattr_dfcol()
    class scd_out:
        pass
    @getattr_dfcol()
    class scd_out_cum:
        pass
    @getattr_dfcol()
    class bal_strt:
        pass
    @getattr_dfcol()
    class amt_in:
        pass
    @getattr_dfcol()
    class amt_in_cum:
        pass
    @getattr_dfcol()
    class amt_out:
        pass
    @getattr_dfcol()
    class amt_out_cum:
        pass
    @getattr_dfcol()
    class bal_end:
        pass
    @getattr_dfcol()
    class rsdl_in_cum:
        pass
    @getattr_dfcol()
    class rsdl_out_cum:
        pass
        
    def _set_outputfunc(self):
        self.scd_in = self.scd_in(self)
        self.scd_in_cum = self.scd_in_cum(self)
        self.scd_out = self.scd_out(self)
        self.scd_out_cum = self.scd_out_cum(self)
        self.bal_strt = self.bal_strt(self)
        self.amt_in = self.amt_in(self)
        self.amt_in_cum = self.amt_in_cum(self)
        self.amt_out = self.amt_out(self)
        self.amt_out_cum = self.amt_out_cum(self)
        self.bal_end = self.bal_end(self)
        self.rsdl_in_cum = self.rsdl_in_cum(self)
        self.rsdl_out_cum = self.rsdl_out_cum(self)
    
    def amt_rqrd_excs(self, idxval, rqrdamt, minunit = 100):
        """
        Calculate the additional amount required in excess of the balance.
        
        Parameters
        ----------
        index : index number
        rqrdamt : int, float, total required amount
        minunit : int, default 100, minimum adjustment unit
        
        Returns
        -------
        amt_rqrd : float
        """
        if isinstance(idxval, int): 
            idxval = self.index[idxval]
        elif isinstance(idxval, str):
            idxval = str_to_date(idxval)
        
        amt_rqrd = max(rqrdamt - self.bal_end[idxval], 0)
        amt_rqrd = round_up(amt_rqrd, -log10(minunit))
        return amt_rqrd
    
    def send(self, idxval, amt, account, note=None):
        """
        Transfer the amount from this account to the opponent account.
        
        Parameters
        ----------
        index : index
        amt : int, float, amount to transfer
        account : account
        note : str, default None
        
        Returns
        -------
        None
        """
        if isinstance(idxval, int): 
            idxval = self.index[idxval]
        elif isinstance(idxval, str):
            idxval = str_to_date(idxval)
            
        if 'title' in account.__dict__:
            account_title = account.title
        else:
            account_title = None
        self.subamt(idxval, amt, account_title, note)
        
        if 'title' in self.__dict__:
            self_title = self.title
        else:
            self_title = None
        account.addamt(idxval, amt, self_title, note)
        
    def __repr__(self):
        """Return a string representation for this object."""
        if len(self.df) > 5:
            repr = f"{self.df.head(2)}\n...\n{self.df.tail(2)}"
        else:
            repr = f"{self.df}"
        if 'byname' in self.__dict__:
            repr = f"<{self.byname}>\n" + repr
        if 'title' in self.__dict__:
            repr = f"<{self.title}>\n" + repr
        return repr
        

class set_initial_account_decorator():
    def __call__(self, cls):
        def init(self, sprcls, title, byname=None, idx=None, crit="addscd",
                 **kwargs):
            self.sprcls = sprcls
            if "_dct" not in sprcls.__dict__:
                sprcls._dct = {}
                sprcls.dct = sprcls._dct
                
                @property
                def mrg(self):
                    return Merge(sprcls.dct)
                sprcls.mrg = mrg
            
            if title not in sprcls._dct:
                acc = Account(idx, title)
                sprcls._dct[title] = acc
                setattr(sprcls, title, acc)
                self.acc = acc
                
                acc.byname = byname
                
                self.crit = crit
                self.kwargs = kwargs
                
            for key, item in kwargs.items():
                setattr(self, key, item)
            #self.istnc = getattr(self.sprcls, title)
            self._initialize()
        cls.__init__ = init
        
        return cls

        
@set_initial_account_decorator()
class set_acc:
    """
    Set initial account. 
    Create and initiaize an account.
    
    Parameters
    ----------
    sprcls : an account on a higher level
    title : a name of the account
    byname : a second name of the account
    idx : index, default None
    crit : str, default "addscd"
    
    Returns
    -------
    
    init(self, sprcls, title, byname=None, idx=None, crit="addscd",
         **kwargs: None)
    """
    def _initialize(self):
        pass
        
@set_initial_account_decorator()
class set_once:
    """
    Set initial account. 
    Create and initiaize an account.
    Input a data(one scheduel, one amount).
    
    Parameters
    ----------
    sprcls : an account on a higher level
    title : a name of the account
    byname : a second name of the account
    idx : index, default None
    crit : str, default "addscd" or "subscd"
    scdidx : a schedule index
    amtttl : an amount to input on
    
    Returns
    -------
    
    init(self, sprcls, title, byname=None, idx=None, crit="addscd",
         **kwargs: amtttl, scdidx)
    """
    def _initialize(self):
        if self.crit == "addscd":
            self.acc.addscd(self.scdidx, self.amtttl)
        elif self.crit == "subscd":
            self.acc.subscd(self.scdidx, self.amtttl)
            
@set_initial_account_decorator()
class set_scd:
    """
    Set initial account. 
    Create and initiaize an account.
    Input a data(scheduel list, amount list).
    
    Parameters
    ----------
    sprcls : an account on a higher level
    title : a name of the account
    byname : a second name of the account
    idx : index, default None
    crit : str, default "addscd" or "subscd"
    scdidx : a schedule index list
    scdamt : an amount list to input on
    
    Returns
    -------
    
    init(self, sprcls, title, byname=None, idx=None, crit="addscd", 
         **kwargs: scdamt, scdidx)
    """
    def _initialize(self):
        if self.crit == "addscd":
            self.acc.addscd(self.scdidx, self.scdamt)
        elif self.crit == "subscd":
            self.acc.subscd(self.scdidx, self.scdamt)
        
@set_initial_account_decorator()
class set_rate:
    """
    Set initial account. 
    Create and initiaize an account.
    Input a data(scheduel list, amount list).
    
    Parameters
    ----------
    sprcls : an account on a higher level
    title : a name of the account
    byname : a second name of the account
    idx : index, default None
    crit : str, default "addscd" or "subscd"
    scdidx : a schedule index
    amt : an amount to apply a rate on
    rate : a rate to apply to an amt
    
    Returns
    -------
    init(self, sprcls, title, byname=None, idx=None, crit="addscd", 
         **kwargs: amt, rate, scdidx)
    """
    def _initialize(self):
        scdamt = self.amt * self.rate
        if self.crit == "addscd":
            self.acc.addscd(self.scdidx, scdamt)
        elif self.crit == "subscd":
            self.acc.subscd(self.scdidx, scdamt)      

        
class Merge(object):
    def __init__(self, dct:dict):
        self._dct = dct
        self._set_mainidx()
        self._set_outputfunc()
        
    @property
    def _df(self):
        dflst = [self._adjust_idx(item._df) for key, item in self._dct.items()]
        return sum(dflst)
        
    @property
    def df(self):
        dflst = [self._adjust_idx(item.df) for key, item in self._dct.items()]
        return sum(dflst)
        
    @property
    def dct(self):
        return self._dct
        
    @property
    def title(self):
        dfdct = Series({key: item.title for key, item in self._dct.items()})
        return dfdct
        
    def __getattr__(self, attr):
        return [item.__dict__[attr] for key, item in self._dct.items()]
        
    def _set_mainidx(self):
        mainidx = []
        for key, item in self._dct.items():
            if len(item.index) > len(mainidx):
                mainidx = item.index
        self.index = DateIndex._simple_new(mainidx)
        #self.index = self.index.arr
        
    def _adjust_idx(self, tmpdf):
        if len(tmpdf.index) < len(self.index):
            return DataFrame(tmpdf, index=self.index).fillna(0)
        return tmpdf
        
    class getattr_dfcol:
        """
        Decorator
        Get a class name and use the class name as the column of dataframe.
        """
        def __call__(self, cls):
            def init(self, spristnc):
                self.spristnc = spristnc
                self.colname = cls.__name__
            cls.__init__ = init
            
            def getitem(self, val):
                """
                If val is an integer, return the data which is in index[val].
                If val is a date, return the data which is on the date.
                If val is a string, get the date data and return the data 
                    which is on the date.
                
                Parameters
                ----------
                val: int, slice, date, date like string, 
                    ex) 0, 1:3, datetime.date(2021, 4, 30), "2021-04", "2021-04-30"
                
                Return
                ------
                data from dataframe
                Array of data from dataframe
                
                Examples
                --------
                >>> idx = DateIndex("2021.01", "2021.12")
                >>> acc = Account(idx, "loan")
                >>> acc.addscd(idx[0], 1000)
                >>> acc.subscd(idx[5], 800)
                >>> acc.addamt(idx[1], 500, "acc_oprtg", "amount acc oprtg")
                >>> acc.scd_in[idx[3]]
                    0.0
                >>> acc.scd_in[3]
                    0.0
                >>> acc.scd_in[0:3]
                    2021-01-31    1000.0
                    2021-02-28       0.0
                    2021-03-31       0.0
                    Name: scd_in, dtype: float64
                >>> acc.scd_in[datetime.date(2021, 3, 31)]
                    0.0
                >>> acc.scd_in["2021.04"]
                    2021-04-30    0.0
                    Name: scd_in, dtype: float64
                >>> acc.scd_in["2021"]
                    2021-01-31    1000.0
                    2021-02-28       0.0
                    2021-03-31       0.0
                    2021-04-30       0.0
                    2021-05-31       0.0
                    2021-06-30       0.0
                    2021-07-31       0.0
                    2021-08-31       0.0
                    2021-09-30       0.0
                    2021-10-31       0.0
                    2021-11-30       0.0
                    Name: scd_in, dtype: float64
                """
                if isinstance(val, date):
                    return self.spristnc._df.loc[val, self.colname]
                val = self.spristnc.index[val]
                return self.spristnc._df.loc[val, self.colname]
            cls.__getitem__ = getitem
            
            return cls
        
    @getattr_dfcol()
    class scd_in:
        pass
    @getattr_dfcol()
    class scd_in_cum:
        pass
    @getattr_dfcol()
    class scd_out:
        pass
    @getattr_dfcol()
    class scd_out_cum:
        pass
    @getattr_dfcol()
    class bal_strt:
        pass
    @getattr_dfcol()
    class amt_in:
        pass
    @getattr_dfcol()
    class amt_in_cum:
        pass
    @getattr_dfcol()
    class amt_out:
        pass
    @getattr_dfcol()
    class amt_out_cum:
        pass
    @getattr_dfcol()
    class bal_end:
        pass
    @getattr_dfcol()
    class rsdl_in_cum:
        pass
    @getattr_dfcol()
    class rsdl_out_cum:
        pass
        
    def _set_outputfunc(self):
        self.scd_in = self.scd_in(self)
        self.scd_in_cum = self.scd_in_cum(self)
        self.scd_out = self.scd_out(self)
        self.scd_out_cum = self.scd_out_cum(self)
        self.bal_strt = self.bal_strt(self)
        self.amt_in = self.amt_in(self)
        self.amt_in_cum = self.amt_in_cum(self)
        self.amt_out = self.amt_out(self)
        self.amt_out_cum = self.amt_out_cum(self)
        self.bal_end = self.bal_end(self)
        self.rsdl_in_cum = self.rsdl_in_cum(self)
        self.rsdl_out_cum = self.rsdl_out_cum(self)

    def __repr__(self):
        """Return a string representation for this object."""
        repr_key = []
        for key in self.dct.keys():
            repr_key.append(key)
        
        if len(self.df) > 5:
            repr_df = f"{self.df.head(2)}\n...\n{self.df.tail(2)}"
        else:
            repr_df = f"{self.df}"
        repr = f"Accounts: {repr_key}\n" + repr_df
        return repr
        
        
        
        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Written by KP_Hong <kphong16@daum.net>
2021, KP_Hong

General functions for cafle.

Attributes
----------
PY : the area unit used in Korea

Modules
-------
Area : transform between m2 and PY units. 
"""

import math
from datetime import datetime
from functools import wraps
from pandas.api.types import is_numeric_dtype

__all__ = ['PY', 'Area', 'EmptyClass', 'is_iterable', 'limited', 'rounding', 
           'print_rounding', 'round_up', 'log10']

PY = 1/3.305785
# PY("pyung") is the area unit used in Korea.

class Area:
    """
    Transform between m2 and PY units.
    
    Parameters
    ----------
    m2 : int, float, default None
    py : int, float, default None
    roundunit : int, default 2
    
    Attributes
    ----------
    area : tuple
        return a tuple of m2, py.
    m2 : float
        value of m2 unit
    py : float
        value of py unit
        
    Examples
    --------
    >>> ar = Area(1000)
    >>> ar.area
        (1000, 302.5)
    >>> ar.py
        302.5
    >>> ar.m2
        1000
    """
    def __init__(self, m2=None, py=None, roundunit=2):
        self._m2 = m2
        self._py = py
        self._roundunit = roundunit
        self._intlz()
        
    def _intlz(self):
        if all([self._m2 is not None, self._py is None]):
            self._py = round(self._m2 * PY, self._roundunit)
        if all([self._m2 is None, self._py is not None]):
            self._m2 = round(self._py / PY, self._roundunit)
            
    def __repr__(self):
        repr_tmp = f"{self.m2:,.0f}m2 ({self.py:,.0f}py)"
        return repr_tmp
    
    @property
    def area(self):
        return (self._m2, self._py)
        
    @property
    def m2(self):
        return self._m2
        
    @property
    def py(self):
        return self._py
        
        
class EmptyClass:
    def __init__(self):
        pass
        
    def __getattr__(self, attr):
        return self.__dict__[attr]
        
    def __getitem__(self, key):
        return self.__dict__[key]

    @property
    def dict(self):
        return self.__dict__

    @property
    def keys(self):
        return self.__dict__.keys()
        

def is_scalar(data):
    if isinstance(data, (str, int, float)):
        return True
    else:
        return False


def is_iterable(data):
    if type(data) == str:
        return False
    try:
        _ = iter(data)
        return True
    except TypeError:
        return False
        
        
def limited(val, upper=None, lower=None):
    """
    Adjust the input value between "upper" and "lower" values.
    
    Parameters
    ----------
    val : int, float
        input value
    upper : int, float, list, tuple
        the criteria for distinguishing upper values
    lower : int, float, list, tuple
        the criteria for distinguishing lower values
    
    Return
    ------
    Adjusted value between upper values and lower values.
        
    Examples
    --------
    >>> limited(100, upper=90, lower=50)
        90
    >>> limited(30, lower=[10, 40])
        40
    """

    tmp_val = val
    
    if upper:
        if is_iterable(upper):
            for val_lmt in upper:
                tmp_val = min(tmp_val, val_lmt)
        else:
            tmp_val = min(tmp_val, upper)
            
    if lower:
        if is_iterable(lower):
            for val_lmt in lower:
                tmp_val = max(tmp_val, val_lmt)
        else:
            tmp_val = max(tmp_val, lower)
    return tmp_val
    
    
def rounding(df, rate=None):
    rslt_df = df.copy()
    for key, item in rslt_df.items():
        if rate is not None and key in rate:
            rslt_df[key] = rslt_df[key] * 100
        if all([isinstance(val, (int, float)) for val in rslt_df[key]]):
            rslt_df[key] = rslt_df[key].fillna(0).apply(lambda x: f"{x:,.0f}")
    return rslt_df
    
    
def print_rounding(df):
    """
    Apply a separator to the number and round the number.
    
    Parameters
    ----------
    df : DataFrame
    
    Return
    ------
    DataFrame rounded and applied a separator to the number
        
    Examples
    --------
    >>> df = DataFrame({'a':[100000,     200000    ], 
                        'b':[100000.123, 200000.321], 
                        'c':['abc',      'cde'     ]})
    >>> print_rounding(df)
                 a        b    c
        0  100,000  100,000  abc
        1  200,000  200,000  cde
    """
    rslt_df = df.copy()
    for key, item in rslt_df.items():
        if is_numeric_dtype(item):
            rslt_df[key] = rslt_df[key].fillna(0).apply(lambda x: f"{x:,.0f}")
    return rslt_df
        
        
def round_up(number:float, decimals:int=2):
    """
    Return a value rounded up to a specific number of decimal places.
    
    Parameters
    ----------
    number : int, float
    decimals : int
    
    Return
    ------
    float
        
    Examples
    --------
    >>> round_up(123.1231, 1)
        123.2
    """
    if not isinstance(decimals, int):
        raise TypeError("decimal places must be an integer")
    elif decimals == 0:
        return math.ceil(number)
        
    factor = 10 ** decimals
    return math.ceil(number * factor) / factor

    
def log10(val):
    tmpval = 0
    while True:
        val = val / 10
        if val > 0.9:
            tmpval += 1
        else:
            return tmpval
    

# Decorator
def listwrapper(func):
    @wraps(func)
    def wrapped(self, *args, **kwargs):
        is_iter = True
        for arg in args:
            if is_iterable(arg) is False:
                is_iter = False
        for item in kwargs.values():
            if is_iterable(item) is False:
                is_iter = False
                
        if is_iter is True:
            ilen = 0
            for arg in args:
                if len(arg) > ilen:
                    ilen = len(arg)
            for item in kwargs.values():
                if len(item) > ilen:
                    ilen = len(item)
            
            for i in range(ilen):
                new_args = []
                new_kwargs = {}
                
                for val in args:
                    new_args = new_args + [val[i]]
                new_args = tuple(new_args)
                
                for key, item in kwargs.items():
                    new_kwargs[key] = item 
                
                func(self, *new_args, **new_kwargs)
        else:
            new_args = args
            new_kwargs = kwargs
            func(self, *new_args, **new_kwargs)
    return wrapped
    


"""
# Decorator
def listwrapper(func):
    @wraps(func)
    def wrapped(self, *args):
        is_iter = True
        for arg in args:
            if is_iterable(arg) is False:
                is_iter = False
        if is_iter is True:
            ilen = len(args[0])
            for i in range(ilen):
                new_args = []
                for val in args:
                    new_args = new_args + [val[i]]
                new_args = tuple(new_args)
                func(self, *new_args)
        else:
            new_args = args
            func(self, *new_args)
    return wrapped
"""
    
    
    
    
    
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Dec 19 17:41:35 2021

@author: KP_Hong
"""

import xlsxwriter
from datetime import datetime
from pandas import DataFrame
from cafle.genfunc import is_iterable, is_scalar
from collections import namedtuple

__all__ = ['Write', 'WriteWS', 'Cell']

class Cell:
    def __new__(cls, row, col):
        cell = namedtuple("Cell", ["row", "col"])
        return cell(row, col)


def depth_dct(dct):
    tlen = 0
    for item in dct.values():
        if type(item) is dict:
            tlen = max(depth_dct(item), tlen)
        elif is_iterable(item):
            tlen = max(len(item), tlen)
        else:
            tlen = max(1, tlen)
    return 1 + tlen

class Write(object):
    def __init__(self, file_adrs):
        self.file_adrs = file_adrs
        self.wb = xlsxwriter.Workbook(self.file_adrs)
        self.ws = {}
        
    # Close workbook
    def close(self):
        self.wb.close()
        return True
        
    # Add worksheet
    def add_ws(self, wsname):
        self.ws[wsname] = self.wb.add_worksheet(wsname)
        return self.ws[wsname]
        
    # Write val
    def write(self, cell, val, fmt=None, wsname=None, drtn='row'):
        self.rtnws(wsname).write(cell.row, cell.col, val, fmt)
        row = cell.row
        col = cell.col
        
        if drtn == "row":
            row += 1
        elif drtn == "col":
            col += 1
        return Cell(row, col)
    
    # Write row
    def write_row(self, cell, data, fmt=None, wsname=None, drtn='row'):
        row = cell.row
        col = cell.col
        dlen = 0
        
        if not is_iterable(fmt):
            fmt = [fmt] * len(data)
                
        if is_iterable(data[0]):
            for val, _fmt in zip(data, fmt):
                self.rtnws(wsname).write_row(row, col, val, _fmt)
                dlen = max(len(val), dlen)
                row += 1
        elif not is_iterable(data[0]):
            _col = col
            for val, _fmt in zip(data, fmt):
                self.rtnws(wsname).write(row, _col, val, _fmt)
                _col += 1
            dlen = max(len(data), dlen)
            row += 1
            
        if drtn=='col':
            row = cell.row
            col = cell.col + dlen
        return Cell(row, col)
        
    # Write column
    def write_col(self, cell, data, fmt=None, wsname=None, drtn='col'):
        row = cell.row
        col = cell.col
        dlen = 0
        
        if not is_iterable(fmt):
            fmt = [fmt] * len(data)
        
        if is_iterable(data[0]):
            for val, _fmt in zip(data, fmt):
                self.rtnws(wsname).write_column(row, col, val, _fmt)
                dlen = max(len(val), dlen)
                col += 1
        elif not is_iterable(data[0]):
            _row = row
            for val, _fmt in zip(data, fmt):
                self.rtnws(wsname).write(_row, col, val, _fmt)
                _row += 1
            dlen = max(len(data), dlen)
            col += 1
            
        if drtn == 'row':
            row = cell.row + dlen
            col = cell.col
        return Cell(row, col)
    
    # Write dictionary column
    def write_dct_col(self, cell, data, fmt=None, wsname=None, drtn='col'):
        row = cell.row
        col = cell.col
        fmtlst = []
        
        if is_iterable(fmt):
            if len(fmt) == 1:
                fmtlst = fmt * 2
            else:
                fmtlst = fmt
        else:
            fmtlst = [fmt] * 2
        
        for key, item in data.items():
            if type(item) is not dict:
                self.rtnws(wsname).write(row, col, key, fmtlst[0])
                if is_iterable(item):
                    self.rtnws(wsname).write_column(row+1, col, item, fmtlst[1])
                else:
                    self.rtnws(wsname).write(row+1, col, item, fmtlst[1])
                col += 1
            if type(item) is dict:
                self.rtnws(wsname).write(row, col, key, fmtlst[0])
                tcell = self.write_dct_col(Cell(row+1, col), item, fmtlst[1:], wsname, drtn='col')
                row = tcell.row - 1
                col = tcell.col
            
        if drtn == 'row':
            row = cell.row + depth_dct(data)
            col = cell.col
        return Cell(row, col)
        
            
    # Write Dictionary
    def write_dct_row(self, cell, data, fmt=None, wsname=None, drtn='row'):
        row = cell.row
        col = cell.col
        fmtlst = []
        
        if is_iterable(fmt):
            if len(fmt) == 1:
                fmtlst = fmt * 2
            else:
                fmtlst = fmt
        else:
            fmtlst = [fmt] * 2
            
        for key, item in data.items():
            if type(item) is not dict:
                self.rtnws(wsname).write(row, col, key, fmtlst[0])
                if is_iterable(item):
                    self.rtnws(wsname).write_row(row, col+1, item, fmtlst[1])
                else:
                    self.rtnws(wsname).write(row, col+1, item, fmtlst[1])
                row += 1
            if type(item) is dict:
                self.rtnws(wsname).write(row, col, key, fmtlst[0])
                tcell = self.write_dct_row(Cell(row, col+1), item, fmtlst[1:], wsname, drtn='row')
                row = tcell.row
                col = tcell.col - 1
        
        if drtn == 'col':
            row = cell.row
            col = cell.col + depth_dct(data)
        return Cell(row, col)
        
    # Write DataFrame
    def write_df_col(self, cell, data, fmt=None, wsname=None, drtn='row'):
        dfdict = data.to_dict('split')
        row = cell.row
        col = cell.col
        
        dfcollen = len(dfdict['columns'])
        dfidxlen = len(dfdict['index'])
        
        if isinstance(dfdict['columns'][0], tuple):
            dfcolno = len(dfdict['columns'][0])
        else:
            dfcolno = 1
            
        if isinstance(dfdict['index'][0], tuple):
            dfidxno = len(dfdict['index'][0])
        else:
            dfidxno = 1
        
        if not is_iterable(fmt):
            fmtlst = [fmt] * dfcollen
        else:
            fmtlst = fmt
            
        if dfcolno == 1:
            self.write_row(Cell(row, col+dfidxno), dfdict['columns'], None, wsname)
        elif dfcolno > 1:
            _row = row
            _col = col + dfidxno
            for valtpl in dfdict['columns']:
                self.write_col(Cell(_row, _col), valtpl, None, wsname)
                _col += 1
        
        if dfidxno == 1:
            self.write_col(Cell(row+dfcolno, col), dfdict['index'], None, wsname)
        elif dfidxno > 1:
            _row = row + dfcolno
            _col = col
            for valtpl in dfdict['index']:
                self.write_row(Cell(_row, _col), valtpl, None, wsname)
                _row += 1
        
        _row = row+dfcolno
        _col = col+dfidxno
        for valtpl in dfdict['data']:
            self.write_row(Cell(_row, _col), valtpl, fmtlst, wsname)
            _row += 1
            
        if drtn == 'col':
            row = cell.row
            col = cell.col + dfidxno + dfcollen
        elif drtn == 'row':
            row = cell.row + dfcolno + dfidxlen
            col = cell.col
        return Cell(row, col)
        
        
    # Make Dictionary of Loan
    def dct_loan(self, ln):
        tmpdct = {}
        tmpdct["Notional_" + ln.title] = \
            {"scd_out": ln.ntnl._df.scd_out,
             "scd_in": ln.ntnl._df.scd_in,
             "amt_out" : ln.ntnl._df.amt_out,
             "amt_in" : ln.ntnl._df.amt_in,
             "bal_end" : ln.ntnl._df.bal_end}
        
        if 'IR' in ln.keys:
            tmpdct["IR_" + ln.title] = \
                {"amt_in" : ln.IR._df.amt_in,
                 "bal_end" : ln.IR._df.bal_end}
        if 'fee' in ln.keys:
            tmpdct["Fee_" + ln.title] = \
                {"amt_in" : ln.fee._df.amt_in,
                 "bal_end" : ln.fee._df.bal_end}
        if 'fob' in ln.keys:
            tmpdct["Fob_" + ln.title] = \
                {"amt_in" : ln.fob._df.amt_in,
                 "bal_end" : ln.fob._df.bal_end}
        return tmpdct

    # Make Dictionary of Loan
    def dctprt_loan(self, ln):
        tmpdct = {}
        tmpdct["Notional_" + ln.title] = \
            {"인출한도": ln.ntnl._df.scd_out,
             "상환예정": ln.ntnl._df.scd_in,
             "인출금액": ln.ntnl._df.amt_out,
             "상환금액": ln.ntnl._df.amt_in,
             "대출잔액": ln.ntnl._df.bal_end}

        if 'IR' in ln.keys:
            tmpdct["IR_" + ln.title] = \
                {"이자금액": ln.IR._df.amt_in,
                 "누적이자": ln.IR._df.bal_end}
        if 'fee' in ln.keys:
            tmpdct["Fee_" + ln.title] = \
                {"수수료금액": ln.fee._df.amt_in,
                 "누적수수료": ln.fee._df.bal_end}
        if 'fob' in ln.keys:
            tmpdct["미인출_" + ln.title] = \
                {"미인출수수료": ln.fob._df.amt_in,
                 "누적미인출": ln.fob._df.bal_end}
        return tmpdct

    # Extend List
    def extndlst(self, lst, *arg):
        for val in arg:
            lst.extend(val)
        return lst
    
    # Extend Dictionary
    def extnddct(self, dct, *arg):
        for val in arg:
            dct = dict(dct, **val)
        return dct
    
    # Return worksheet
    def rtnws(self, wsname):
        if isinstance(wsname, str):
            return self.ws[wsname]
        elif isinstance(wsname, xlsxwriter.worksheet.Worksheet):
            return wsname
        
    # Text Format
    def fmtnum(self, fmt, **kwargs):
        """
        example of fmt: 'yyyy-mm-dd', '#,##0', '#,##0.0', '0.0%'
        kwargs example: bold=True
        """
        tmpdct = dict({'num_format': fmt, **kwargs})
        return self.wb.add_format(tmpdct)
    
    @property
    def nml(self):
        return self.wb.add_format({'bold': False})
    
    @property
    def bold(self):
        return self.wb.add_format({'bold': True})
        
    @property
    def num(self):
        return self.wb.add_format({'num_format': '#,##0'})
        
    @property
    def numb(self):
        return self.wb.add_format({'num_format': '#,##0', 'bold': True})
        
    @property
    def pct(self):
        return self.wb.add_format({'num_format': '0.0%'})
        
    @property
    def pct2(self):
        return self.wb.add_format({'num_format': '0.00%'})
        
    @property
    def date(self):
        return self.wb.add_format({'num_format': 'yyyy-mm-dd'})

    @property
    def month(self):
        return self.wb.add_format({'num_format': '#,##0"개월"'})
    
    @property
    def now(self):
        return datetime.now().strftime('%Y.%m.%d %H:%M:%S')


class WriteWS(Write):
    def __init__(self, ws, cell):
        self.ws = ws
        self.cell = cell

    def __call__(self, data, fmt=None, valdrtn='row', drtn='row'):
        if not is_iterable(data):
            self.cell = self.write(self.cell, data, fmt, self.ws, drtn)
        elif isinstance(data, DataFrame):
            if valdrtn == 'row':
                raise ValueError("function for row option is not written.")
            elif valdrtn == 'col':
                self.cell = self.write_df_col(self.cell, data, fmt, self.ws, drtn)
        elif is_iterable(data) and not isinstance(data, dict):
            if valdrtn == 'row':
                self.cell = self.write_row(self.cell, data, fmt, self.ws, drtn)
            elif valdrtn == 'col':
                self.cell = self.write_col(self.cell, data, fmt, self.ws, drtn)
        elif isinstance(data, dict):
            if valdrtn == 'row':
                self.cell = self.write_dct_row(self.cell, data, fmt, self.ws, drtn)
            if valdrtn == 'col':
                self.cell = self.write_dct_col(self.cell, data, fmt, self.ws, drtn)
        return self.cell
    
    def setcell(self, row=None, col=None):
        if isinstance(row, tuple):
            self.cell = row
            return None
        
        _row = self.cell.row
        _col = self.cell.col
        if row is not None:
            _row = row
        if col is not None:
            _col = col
        self.cell = Cell(_row, _col)
        return None
    
    def nextcell(self, no=1, drtn='row'):
        _row = self.cell.row
        _col = self.cell.col
        if drtn=='row':
            _row += no
        elif drtn=='col':
            _col += no
        self.cell = Cell(_row, _col)
        return self.cell
        
    def dctdatalen(self, dct):
        _len = 0
        for key, item in dct.items():
            if is_iterable(item):
                _len = max(_len, len(item))
        return _len


        
        
        
        
        
        
        
        
        
        
        
    
    
    


from types import FunctionType
from .account import (
    Account,
    Merge,
)

__all__ = ['SetAccount']

#### Set Account ####
class SetAccount:
    def __init__(self, title=None, byname=None):
        self.title = title
        self.byname = byname
        self._dct = {}

    @property
    def keys(self):
        return self.__dict__.keys()

    def __getitem__(self, key):
        return self.__dict__[key]

    def set_account(self, title, byname=None, tag=None):
        _acc = Account(title=title, byname=byname)

        setattr(self, title, _acc)
        self._dct[title] = _acc

        if isinstance(tag, str):
            self.tag_account(_acc, tag)
        elif isinstance(tag, list):
            for tagval in tag:
                self.tag_account(_acc, tagval)
        return _acc

    def get_account(self, _acc):
        title = _acc.title
        if 'tag' in _acc.__dict__:
            tag = _acc.tag
        else:
            tag = None
        setattr(self, title, _acc)
        self._dct[title] = _acc
        if isinstance(tag, str):
            self.tag_account(_acc, tag)
        elif isinstance(tag, list):
            for tagval in tag:
                self.tag_account(_acc, tagval)
        return _acc

    def tag_account(self, acc, tag):
        if tag in self.__dict__.keys():
            if isinstance(getattr(self, tag), dict):
                pass
            else:
                raise NameError("The tag name is already existing in account keys.")
        else:
            setattr(self, tag, {})
        getattr(self, tag)[acc.title] = acc

    def get_item(self, valdct=None, item='amt_ttl', func='sum'):
        if valdct is None:
            valdct = self._dct
        if isinstance(valdct, str):
            valdct = getattr(self, valdct)
        itemlst = []
        for tmpval in valdct.values():
            try:
                itemval = getattr(tmpval, item)
                itemlst.append(itemval)
            except:
                pass
        if isinstance(func, str):
            return eval(f"{func}({str(itemlst)})")
        elif isinstance(func, FunctionType):
            return func(itemlst)

    def mrg_account(self, valdct=None):
        if valdct is None:
            valdct = self._dct
        if isinstance(valdct, str):
            valdct = getattr(self, valdct)
        return Merge(valdct)

    @property
    def dct(self):
        if len(self._dct) == 0:
            return None
        return self._dct


import os
import pandas as pd

__all__ = ['read_standard_process_rate_table']

def read_standard_process_rate_table(colno=None, tolist=False):
    DIRECTORY = '/'.join(os.path.abspath(__file__).split('/')[:-1])
    filename = "/standard_data/standard_process_rate_table.csv"
    fileloc = DIRECTORY + filename

    result = pd.read_csv(fileloc, index_col="idx")

    if colno is None:
        return result
    else:
        colno = str(colno)
        result = result[colno].dropna(axis=0)
        if tolist is False:
            return result
        if tolist is True:
            return result.values.tolist()